from logging import getLogger

from sqlalchemy import Connection, inspect
from sqlalchemy.orm import DeclarativeBase, Mapper

from sqla_history.stmt_builder import InsertStmtBuilder

from .context import CurrentUserId, EventId
from .dto import HistoryCreateDTO, Value
from .not_set import NOT_SET, NotSet
from .types_ import UserId
from .utils import utc_now

logger = getLogger(__name__)


class ChangeEventHandler:
    def __init__(
        self,
        entity_name: str,
        stmt_builder: InsertStmtBuilder,
        id_field_name: str = "id",
    ) -> None:
        self._entity_name = entity_name
        self._stmt_builder = stmt_builder
        self._id_field_name = id_field_name

    def __call__(
        self,
        mapper: Mapper,
        connection: Connection,
        target: DeclarativeBase,
    ) -> None:
        if (event_id := EventId.get()) is None:
            logger.info("event_id is None. Changes wasn't tracked")
            return

        state = inspect(target)
        ident = getattr(target, self._id_field_name)
        changed_at = utc_now()

        dtos: list[HistoryCreateDTO] = []
        for attr in mapper.columns:
            field_name = attr.key
            new_value = getattr(target, field_name)
            attribute_state = state.attrs[field_name]
            if not attribute_state.history.deleted:
                continue

            prev_value = attribute_state.history.deleted[0]
            if prev_value == new_value:  # pragma: no cover
                # no cover as the case could not be reproduced
                continue

            dtos.append(
                HistoryCreateDTO(
                    event_id=event_id,
                    entity_id=ident,
                    entity_name=self._entity_name,
                    changed_at=changed_at,
                    field_name=field_name,
                    prev_value=Value(prev_value).model_dump(mode="json"),
                    new_value=Value(new_value).model_dump(mode="json"),
                    user_id=self._get_user_id(),
                )
            )

        if not dtos:
            return

        stmt = self._stmt_builder.build(dtos)
        connection.execute(stmt)

    def _get_user_id(self) -> UserId | None | NotSet:
        return NOT_SET


class WithUserChangeEventHandler(ChangeEventHandler):
    def _get_user_id(self) -> UserId | None | NotSet:
        return CurrentUserId.get()



from contextvars import ContextVar, Token
from uuid import UUID

from .types_ import UserId

current_user_id: ContextVar[UserId | None] = ContextVar("current_user_id", default=None)
"""
A context variable that stores the ID of the user who sent the HTTP request or triggered another action
"""


class CurrentUserId:
    '''Helper class for working with the context variable "current_user_id"'''

    @staticmethod
    def set(value: UserId) -> Token[UserId | None]:
        return current_user_id.set(value)

    @staticmethod
    def get() -> UserId | None:
        return current_user_id.get()

    @staticmethod
    def reset(token: Token[UserId | None]) -> None:
        current_user_id.reset(token)


event_id: ContextVar[UUID | None] = ContextVar("event_id", default=None)
"""
A context variable that stores an event identifier,
by which you can combine changes to fields in database tables (see the `sqla_history.base.BaseHistoryChanges`)
"""


class EventId:
    '''Helper class for working with the context variable "event_id"'''

    @staticmethod
    def set(value: UUID) -> Token[UUID | None]:
        return event_id.set(value)

    @staticmethod
    def get() -> UUID | None:
        return event_id.get()

    @staticmethod
    def reset(token: Token[UUID | None]) -> None:
        event_id.reset(token)



from sqla_history.context import CurrentUserId, EventId

from .base import BaseHistoryChanges
from .dto import HistoryCreateDTO, Value
from .event_handlers import ChangeEventHandler, WithUserChangeEventHandler
from .stmt_builder import BaseInsertStmtBuilder, InsertStmtBuilder

__all__ = [
    "BaseHistoryChanges",
    "BaseInsertStmtBuilder",
    "ChangeEventHandler",
    "CurrentUserId",
    "EventId",
    "HistoryCreateDTO",
    "InsertStmtBuilder",
    "Value",
    "WithUserChangeEventHandler",
]



from datetime import datetime
from typing import Any
from uuid import UUID

from sqlalchemy import JSON, String
from sqlalchemy.orm import Mapped, mapped_column
from uuid_utils.compat import uuid7

from .utils import utc_now


class BaseHistoryChanges:
    """
    The base model to inherit from

    Example:
    ```
    from sqlalchemy.orm import DeclarativeBase


    class Base(DeclarativeBase):...


    class HistoryChanges(BaseHistoryChanges, Base):
        user_id: Mapped[UUID | None] = mapped_column(ForeignKey("user.id", on_delete="SET NULL"))
        entity_id: Mapped[UUID] = mapped_column(index=True)

    ```
    """

    id: Mapped[UUID] = mapped_column(primary_key=True, default=uuid7)
    event_id: Mapped[UUID] = mapped_column(default=uuid7)
    changed_at: Mapped[datetime] = mapped_column(default=utc_now)
    entity_name: Mapped[str] = mapped_column(String(255), index=True)
    field_name: Mapped[str] = mapped_column(String(255))
    prev_value: Mapped[Any | None] = mapped_column(JSON())
    new_value: Mapped[Any | None] = mapped_column(JSON())



from collections.abc import Mapping
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Any
from uuid import UUID

from pydantic import RootModel

from .not_set import NOT_SET, NotSet
from .types_ import EntityId, UserId


class Value(RootModel[Any]):
    pass


@dataclass(frozen=True, slots=True, kw_only=True)
class HistoryCreateDTO:
    event_id: UUID
    entity_id: EntityId
    user_id: UserId | None | NotSet = NOT_SET
    entity_name: str
    changed_at: datetime
    field_name: str
    prev_value: Any
    new_value: Any

    def to_dict(self) -> Mapping[str, Any]:
        data = asdict(self)
        if isinstance(self.user_id, NotSet):
            del data["user_id"]

        return data



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import final

from sqlalchemy import Insert, insert
from sqlalchemy.orm import DeclarativeBase

from .dto import HistoryCreateDTO


class BaseInsertStmtBuilder(ABC):  # pragma: no cover
    def __init__(self, history_model: type[DeclarativeBase]) -> None:
        self._history_model = history_model

    @abstractmethod
    def build(
        self,
        dtos: Sequence[HistoryCreateDTO],
    ) -> Insert:
        pass


@final
class InsertStmtBuilder(BaseInsertStmtBuilder):
    def build(
        self,
        dtos: Sequence[HistoryCreateDTO],
    ) -> Insert:
        return insert(self._history_model).values([dto.to_dict() for dto in dtos])
